#!/usr/bin/env python3

#
# Final Bubble Bobble translation patch tool
# Brad Smith, 2020
# http://rainwarrior.ca
# https://www.patreon.com/posts/40326586
#

import shlex # for split
import struct

JAPAN = True
#JAPAN = False # dump Europe version instead (no patch made)
FILEJP = "Final Bubble Bobble (Japan).sms"
FILEEU = "Bubble Bobble (Europe).sms"
ROMFILE = FILEJP if JAPAN else FILEEU
OROMFILE = FILEEU if JAPAN else FILEJP
ROM = open(ROMFILE,"rb").read()
DUMPFILE = ROMFILE + ".txt"
PATCHFILE = "translated.txt"
PATCHROM = "translated.sms"
PATCHIPS = "translated.ips"

# locations of data in ROM
HINT_BANK = 0x14000 # -0x8000 offset because it's in bank 2
HINT_TABLE = 0x9185
HINT_COUNT = 64
HINT_TEXT_END = 0x9ABB
HINT_LINE_MAX = 28 # maximum characters per line
CUT_BANK = 0x14000 # -0x8000 offset
CUT_START = 0xB150
CUT_END = 0xB3B7
CUT_LENGTH = [] # filled in by dump_cuts()
CUT_COUNT = 25
COPY_START = 0x874D
COPY_END = 0x8766

# character set begins at $138 in VRAM
#
# Europe replaces "あいうえお" with "&'., "
# The EUROPE dump will show "&'*/-" instead.
#
# ; is unreachable (replaced by …)
# ? is unreachable (replaced by ？)
# _ is a duplicated space
# x is not in the set, used here for invalid data
# n is not in the set, used here for newline
#
# 0-Z tiles are stored as 1BPP at ROM $20080
# あ and beyond are at ROM $3EE40
# There seem to be a few extra unused ascii characters at $20000
#
global CHARSET
CHARSET = "01234567" + \
  "89:;_= ?@ABCDEFG" + \
  "HIJKLMNOPQRSTUVW" + \
  "XYZ<あいうえおかきくけこさし" +\
  "すせそたちつてとなにぬねのはひふ" + \
  "へほまみむめもやゆよらりるれろわ" + \
  "をんがぎぐげござじずぜぞだぢづで" + \
  "どばびぶべぼぱぴぷぺぽゃゅょっ～" + \
  "、。！？「」♥…xn"

if not JAPAN:
    CHARSET = CHARSET[0:0x2C] + "&'*/-" + CHARSET[0x31:]

# Based on tile decoding routine run at $55F-$5C2
def byte_to_tile(b):
    if b == 0xFF:
        return 0xB9 # newline n character
    if b > 0xD4: # not part of routine but too high to encode
        return 0xB8 # invalid x character
    if b >= 0x80: # japanese characters starting at $164
        return (b + 0xE4) & 0xFF;
    BSUBS = { # a bunch of 1:1 replacements
        0x20 : 0x46,  # space
        0x21 : 0xB2,  # !
        0x3F : 0xB3,  # ?
        0x5B : 0xB4,  # [ (Japanese quotes)
        0x5D : 0xB5,  # ]
        0x26 : 0xB6,  # & (heart)
        0x3B : 0xB7,  # ; (...)
        0x2C : 0xB0,  # ,
        0x2E : 0xB1,  # .
        0x7E : 0xAF } # ~
    if b in BSUBS:
        return BSUBS[b]
    if b < 0x30 or b >= 0x5B: # out of range becomes space
        return 0x46
    return (b + 0x08) & 0xFF # 0-9 :;<=>?@ A-Z <-
    # note: ; $43 and ? $47 are unreachable
    #       < is an extra space character
    #       > is a space character

chartile = {} # converts character to tile number
tilechar = {} # convers tile number to character
tilebyte = {} # converts tile number to byte code
bytetile = {} # converts byte to tile number
charbyte = {} # converts character to byte

for i in range(len(CHARSET)):
    tile = (i + 0x138) & 0xFF
    char = CHARSET[i]
    chartile[char] = tile
    tilechar[tile] = char
    
for b in range(256):
    t = byte_to_tile(b)
    tilebyte[t] = b
    bytetile[b] = t

for (c,t) in chartile.items():
    if t in tilebyte:
        charbyte[c] = tilebyte[t]
# "transparent" alternate space needed for cutscene word bubbles
charbyte['<'] = 0x3C
# a few characters are one-to-many mappings, override them with best choice
charbyte[' '] = 0x20
charbyte['！'] = 0x21
charbyte['♥'] = 0x26
charbyte['、'] = 0x2C
charbyte['。'] = 0x2E
charbyte['…'] = 0x3B
charbyte['？'] = 0x3F
charbyte['「'] = 0x5B
charbyte['」'] = 0x5D
charbyte['～'] = 0x7E
# duplicate mappings for a few characters
charbyte[','] = charbyte['、']
charbyte['.'] = charbyte['。']
charbyte['!'] = charbyte['！']
charbyte['?'] = charbyte['？']

#
# dumping
#

def diff(cutoff=-1):
    if cutoff < 0:
        cutoff = len(ROM)
    OROM = open(OROMFILE,"rb").read()
    print("differences with %s (%d bytes)" % (OROMFILE,len(OROM)))
    cutoff = min(cutoff, len(ROM), len(OROM))
    i = 0
    while i < cutoff:
        if ROM[i] != OROM[i]:
            i += 1
            continue
        start = i
        while i < cutoff and ROM[i] == OROM[i]:
            i += 1
        print("$%05X: %d bytes" % (start,i-start))

def dump_string(addr):
    s = ""
    while ROM[addr] != 0:
        b = ROM[addr]
        t = bytetile[b]
        c = tilechar[t]
        s += c
        if c == 'x':
            print("invalid: $%05X = $%02X ($%02X)" % (addr,b,t))
        addr += 1
    return s

def dump_hints():
    print("%3d hints from table at $%05X" % (HINT_COUNT,HINT_BANK+HINT_TABLE))
    dt = ""
    for h in range(HINT_COUNT):
        ptr_addr = HINT_BANK+HINT_TABLE+(h*2)
        ptr = ROM[ptr_addr] + (ROM[ptr_addr+1] << 8)
        hint_addr = HINT_BANK + ptr
        dump_addr = hint_addr
        dt += "hint %d # $%04X ($%05X)\n" % (h,ptr,hint_addr)
        s = dump_string(dump_addr)
        dump_addr += len(s) + 1
        dt += s.replace("n","n\n") + "\n\n"
        print ("%3d: $%04X ($%05X) ""%s""" % (h,ptr,hint_addr,s))
    return dt

def dump_cuts():
    print("cutscenes from block at $%05X" % (CUT_BANK+CUT_START))
    dt = ""
    dump_addr = CUT_BANK+CUT_START
    dump_end = CUT_BANK+CUT_END
    count = 0
    while dump_addr < dump_end:
        p0 = [(dump_addr-CUT_BANK),dump_addr]
        p1 = list(ROM[dump_addr:dump_addr+4])
        dt += "cut %3d %02X %02X %02X %02X # $%04X ($%05X)\n" % tuple([count]+p1+p0)
        dump_addr += 4
        s = dump_string(dump_addr)
        dump_addr += len(s) + 1
        dt += s.replace("n","n\n") + "\n\n"
        print ("%3d: $%04X ($%05X) = $%02X $%02X $%02X $%02X ""%s""" % tuple([count]+p0+p1+[s]))
        CUT_LENGTH.append(len(s))
        count += 1
    assert len(CUT_LENGTH) == count
    assert count == CUT_COUNT
    return dt

def dump_copy():
    print("copyright string from $%05X" % (COPY_START))
    s = dump_string(COPY_START)
    dt = "copy # $%06X\n" % COPY_START
    dt += s + "\n\n"
    print('"'+ s + '"')
    return dt

#
# patching
#

def ips_chunk(addr, data):
    b = bytearray()
    b.append((addr      >> 16) & 0xFF)
    b.append((addr      >>  8) & 0xFF)
    b.append((addr      >>  0) & 0xFF)
    b.append((len(data) >>  8) & 0xFF)
    b.append((len(data) >>  0) & 0xFF)
    b += bytearray(data)
    return b

def wipe(prom,start,end):
    for i in range(start,end):
        prom[i] = 0

def write_string(prom,addr,s):
    for i in range(len(s)):
        c = s[i]
        if c not in charbyte:
            print("No byte mapping for character '%s' in '%s'" % (c,s))
        prom[addr+i] = charbyte[c]
    prom[addr+len(s)] = 0
    return len(s)+1

global errorline

def patch():
    # parse patch file
    def check(condition, description): # error check during parsing
        if not condition:
            print (description)
            print ("Error on line %d" % errorline)
            assert False
    print(PATCHFILE)
    lines = open(PATCHFILE,"rt",encoding="utf-8").readlines()
    STRING_COUNT = (HINT_COUNT + CUT_COUNT + 1)
    strings = [None] * STRING_COUNT
    cutbox = [(-1,-1,-1,-1)] * CUT_COUNT
    active_str = -1
    errorline = 0
    for l in lines:
        errorline += 1
        l = l.rstrip("\n")
        # identify command lines
        args = shlex.split(l,comments=True)
        if len(args) < 1: # empty line or comment
            continue
        if args[0] == "hint":
            check(len(args) >= 2, "hint has 1 argument")
            i = int(args[1])
            check(i >= 0 and i < HINT_COUNT, "hint %d out of range (0 to %d)" % (i,HINT_COUNT))
            active_str = i
            check(strings[active_str] == None, "duplicate hint")
            continue
        if args[0] == "cut":
            check(len(args) >= 6, "cut has 5 arguments")
            i = int(args[1])
            check(i >= 0 and i < CUT_COUNT, "cut %d out of range (0 to %d)" % (i,CUT_COUNT))
            active_str = HINT_COUNT + i
            check(strings[active_str] == None, "duplicate cut")
            box = tuple(int(x,16) for x in args[2:6])
            check(min(box)>=0 and max(box)<256, "cut box out of range?")
            check(cutbox[i] == (-1,-1,-1,-1), "duplicate cut")
            cutbox[i] = box
            continue
        if args[0] == "copy":
            active_str = HINT_COUNT + CUT_COUNT
            check(strings[active_str] == None, "duplicate copy")
            continue
        # not a command, just append this line to the active string
        check(active_str >= 0,"no hint/cut/copy before first string?")
        if strings[active_str] == None:
            strings[active_str] = ""
        strings[active_str] += l
    # validate
    for i in range(STRING_COUNT):
        if strings[i] == None: # do all strings exist?
            base = 0
            tname = "hint"
            if i >= HINT_COUNT:
                base = HINT_COUNT
                tname = "cut"
            if i >= HINT_COUNT + CUT_COUNT:
                base = 0
                tname = "copy"
            print("%s %d missing!" % (tname,i-base))
            assert False
    MAX_HINT_SIZE = HINT_TEXT_END - (HINT_TABLE + (HINT_COUNT * 2))
    hint_size = 0
    hint_valid = True
    for i in range(HINT_COUNT):
        s = strings[i]
        hint_size += len(s) + 1
        longest_line = max(map(len,s.split("n")))
        if longest_line > HINT_LINE_MAX:
            print("hint %d has a line too long: %d > %d" % (i,longest_line,HINT_LINE_MAX))
            hint_valid = False
    print("hint text bytes: %4d/%4d" % (hint_size,MAX_HINT_SIZE))
    if hint_size > MAX_HINT_SIZE:
        print("Out of space for hint text! %d bytes over." % (hint_size-MAX_HINT_SIZE))
        assert False
    assert hint_valid
    cut_valid = True
    for i in range(CUT_COUNT):
        l = len(strings[HINT_COUNT+i])
        clex = ""
        if l > CUT_LENGTH[i]:
            clex = " too long!"
            cut_valid = False
        print("cut %2d byte: %3d/%3d%s" % (i,l,CUT_LENGTH[i],clex))
    if not cut_valid:
        assert False
    l = len(strings[HINT_COUNT+CUT_COUNT])
    if l != ((COPY_END - COPY_START) - 1):
        print("copy length wrong: %d != %d" % (l,(COPY_END-COPY_START)-1))
        assert False
    # prepare patched ROM
    prom = bytearray(ROM)
    wipe(prom,HINT_BANK+HINT_TABLE,HINT_BANK+HINT_TEXT_END)
    wipe(prom,CUT_BANK +CUT_START, CUT_BANK +CUT_END      )
    wipe(prom,          COPY_START,          COPY_END     )
    hint_text_addr = HINT_TABLE + (HINT_COUNT * 2)
    for i in range(HINT_COUNT):
        hint_table_pos = HINT_BANK + HINT_TABLE + (i*2)
        prom[hint_table_pos+0] = hint_text_addr & 0xFF
        prom[hint_table_pos+1] = hint_text_addr >> 8
        hint_text_addr += write_string(prom, HINT_BANK+hint_text_addr, strings[i])
    cut_addr = CUT_START
    for i in range(CUT_COUNT):
        box = cutbox[i]
        for j in range(4):
            prom[CUT_BANK+cut_addr+j] = box[j]
        s = strings[HINT_COUNT+i].replace(" ","<") # replace solid blank > with transparent blank <
        write_string(prom, CUT_BANK+cut_addr+4, s)
        cut_addr += 4 + CUT_LENGTH[i] + 1
    write_string(prom, COPY_START, strings[HINT_COUNT+CUT_COUNT])
    print(PATCHROM)
    open(PATCHROM,"wb").write(prom)
    # prepare IPS
    ips  = bytearray(map(ord,['P','A','T','C','H']))
    ips += ips_chunk(COPY_START,           prom[          COPY_START:          COPY_END     ])
    ips += ips_chunk(HINT_BANK+HINT_TABLE, prom[HINT_BANK+HINT_TABLE:HINT_BANK+HINT_TEXT_END])
    ips += ips_chunk(CUT_BANK +CUT_START,  prom[CUT_BANK +CUT_START :CUT_BANK +CUT_END      ])
    ips += bytearray(map(ord,['E','O','F']))
    print(PATCHIPS)
    open(PATCHIPS,"wb").write(ips)

#
# main
#

dt = "# ROM %s (%d bytes)\n\n" % (ROMFILE,len(ROM))
print ("%s (%d bytes)" % (ROMFILE,len(ROM)))
#diff(0x1FA00) # diffs above 1FA00 all seem to be graphics
dt += dump_hints()
dt += dump_cuts()
dt += dump_copy()
print ("dump: %s" % DUMPFILE)
open(DUMPFILE,"wt",encoding="utf-8").write(dt)

if JAPAN:
    print()
    print("Patching...")
    patch()

print("Done.")
